/* callwood-ikebana-csg.inc version 1.1
 * Persistence of Vision Ray Tracer include file
 *
 * Newsgroups: povray.binaries.images
 * From: Shay
 * Subject: CSG Challenge
 * Date: 24 Jul 2013 00:02:02
 * Message: <51ef197a@news.povray.org>
 *
 * No attempt was made to eliminate coincident surfaces; therefore, this
 * implementation may be unsuitable for transparent or translucent textures.
 *
 * Copyright © 2013 Richard Callwood III.  Some rights reserved.
 * This work is licensed under a Creative Commons Attribution-NonCommercial-
 * ShareAlike 3.0 Unported License.  For details, see:
 *   http://creativecommons.org/licenses/by-nc-sa/3.0/deed.en_US
 *
 * Vers  Date         Comments
 * ----  ----         --------
 *       23-jul-2013  Test scene is started.
 *       11-aug-2013  Test scene is completed, except for extended testing.
 *       16-aug-2013  Test scene is debugged.
 *       17-aug-2013  Code is copied and adapted from the test scene and from a
 *                    previously written include file of math routines.
 * 1.0   24-aug-2013  Further debugged.
 * 1.1   23-sep-2013  Cusp is boxed for improved speed.
 */
#ifndef (CallwoodIkebanaCSG_Inc_Temp)
#declare CallwoodIkebanaCSG_Inc_Temp = version;
#version 3.5;

#include "math.inc"
#include "shapes.inc"

/*================================ MAIN MACRO ==================================
 * Arguments (all angles are in degrees):
 *   nPetals   - The number of petals
 *   R         - The outer radius of curvature of the bowl
 *   Thickness - The thickness of the wall
 *   Edge      - An edge size, upon which the minor radii of curvature of
 *               rounded edges will be based
 *   arNeck    - The angular radius of the space between the necks of two
 *               adjacent petals
 *   arPetal   - The angular radius of a petal
 *   phiPetal  - The angular distance from the bottom of the bowl to the center
 *               of a petal
 *   arHole    - The angular radius of each hole on the petals
 *   phiHole   - The angular distance from the center of a petal to the center
 *               of a hole
 *   arCusp    - The angular radius of the concave edge of the cusp of a petal
 *   thetaCusp - The angle between the meridian of a petal and the curve that
 *               forms the cusp
 *
 * Try CallwoodIkebanaCSG (8, 110, 3, 1.2, 7, 18.5, 60, 3.6, 6.3, 25, 30) and
 * experiment from there.
 */
#macro CallwoodIkebanaCSG
( nPetals, R, Thickness, Edge,
  arNeck, arPetal, phiPetal, arHole, phiHole, arCusp, thetaCusp
)
 //-------- general parts and dimensions ----------
  #local rc3_RInner = R - Thickness;
  #local rc3_aMargin = degrees (Edge / (R - Thickness/2));
  #local rc3_sinMargin = sind (rc3_aMargin);
  #local rc3_rOuter_edge = R * rc3_sinMargin / (1 + rc3_sinMargin);
  #local rc3_rInner_edge = rc3_RInner * rc3_sinMargin / (1 - rc3_sinMargin);
  #local rc3_Outer_depth = R - sqrt (R*R - 2 * R * rc3_rOuter_edge);
  #local rc3_Inner_depth =
    rc3_rInner_edge * (R - rc3_Outer_depth) / rc3_rOuter_edge - rc3_RInner;
  #local rc3_aHalf = 180 / nPetals;
  #local rc3_aNext = 360 / nPetals;

  #local rc3_Full_shell = difference
  { sphere { 0, R }
    sphere { 0, rc3_RInner }
  }
  #local rc3_Mid_shell = difference
  { sphere { 0, R - rc3_Outer_depth }
    sphere { 0, rc3_RInner + rc3_Inner_depth }
  }

 //--------------- cusp of petal ------------------
  #local rc3_x_Cusp = transform { rotate <arPetal + arCusp, -thetaCusp, 0> }
  #local rc3_phiCusp = CallwoodIkebanaCSG_SAdjacent
  ( arCusp + arPetal, arCusp + rc3_aMargin, thetaCusp, -1,
    "cusp of the petals"
  );
  #local rc3_wedgeCusp = CallwoodIkebanaCSG_SAngle
  ( arCusp + arPetal, arCusp + rc3_aMargin, rc3_phiCusp
  );
  #local rc3_v_Tmp = CallwoodIkebanaCSG_v_Place_torus
  ( arCusp + rc3_aMargin, R - rc3_rOuter_edge
  );
  #local rc3_Cusp_outer_edge = intersection
  { CallwoodIkebanaCSG_Torus
    ( arCusp + rc3_aMargin, rc3_rOuter_edge, R - rc3_rOuter_edge
    )
    box //using a box instead of plane{x,0} forces more efficient bounding
    { < -(rc3_v_Tmp.x + rc3_rOuter_edge) * sind (rc3_wedgeCusp),
        rc3_v_Tmp.y - rc3_rOuter_edge,
        (rc3_v_Tmp.x - rc3_rOuter_edge) * cosd (rc3_wedgeCusp)
      >,
      <0, rc3_v_Tmp.y + rc3_rOuter_edge, rc3_v_Tmp.x + rc3_rOuter_edge>
    }
    plane { -x, 0 rotate -rc3_wedgeCusp * y }
    transform { rc3_x_Cusp }
  }
  #local rc3_v_Tmp = CallwoodIkebanaCSG_v_Place_torus
  ( arCusp + rc3_aMargin, rc3_RInner + rc3_rInner_edge
  );
  #local rc3_Cusp_inner_edge = intersection
  { CallwoodIkebanaCSG_Torus
    ( arCusp + rc3_aMargin, rc3_rInner_edge, rc3_RInner + rc3_rInner_edge
    )
    box //using a box instead of plane{x,0} forces more efficient bounding
    { < -(rc3_v_Tmp.x + rc3_rInner_edge) * sind (rc3_wedgeCusp),
        rc3_v_Tmp.y - rc3_rInner_edge,
        (rc3_v_Tmp.x - rc3_rInner_edge) * cosd (rc3_wedgeCusp)
      >,
      <0, rc3_v_Tmp.y + rc3_rInner_edge, rc3_v_Tmp.x + rc3_rInner_edge>
    }
    plane { -x, 0 rotate -rc3_wedgeCusp * y }
    transform { rc3_x_Cusp }
  }
  #local rc3_Cusp_clip_top = plane
  { -x, 0
    rotate -rc3_wedgeCusp * y
    transform { rc3_x_Cusp }
  }
  #local rc3_Cusp_clip = intersection
  { plane { x, 0 rotate -thetaCusp * y }
    plane { -x, 0 rotate thetaCusp * y }
    union
    { object { rc3_Cusp_clip_top }
      object { rc3_Cusp_clip_top scale <-1, 1, 1> }
    }
    CallwoodIkebanaCSG_Cone
    ( rc3_RInner, R, (arPetal + arCusp) * cosd (thetaCusp)
    )
    object
    { CallwoodIkebanaCSG_Cone (0, R + 1, arPetal - rc3_aMargin)
      inverse
    }
  }

 //--------------- hole in petal ------------------
  #local rc3_Hole = object
  { CallwoodIkebanaCSG_Cone (rc3_RInner - 1, R + 1, arHole)
    rotate <phiHole, 0, 0>
  }

 //---------------- rim of hole -------------------
  #local rc3_rRim = Edge / 2;
  #local rc3_ROuter_rim = sqrt (pow (R, 2) + pow (rc3_rRim, 2));
  #local rc3_RInner_rim = sqrt (pow (rc3_RInner, 2) + pow (rc3_rRim, 2));
  #local rc3_arOuter_rim = arHole + atan2d (rc3_rRim, R);
  #local rc3_arInner_rim = arHole + atan2d (rc3_rRim, rc3_RInner);
  #local rc3_arOuter_rim_cutout = CallwoodIkebanaCSG_Angle
  ( rc3_ROuter_rim, R + rc3_rRim, rc3_rRim * 2
  ) + rc3_arOuter_rim;
  #local rc3_arInner_rim_cutout = CallwoodIkebanaCSG_Angle
  ( rc3_RInner_rim, rc3_RInner - rc3_rRim, rc3_rRim * 2
  ) + rc3_arInner_rim;
  #local rc3_ROuter_rim_blob = CallwoodIkebanaCSG_Rim_blob_R
  ( rc3_ROuter_rim, rc3_arOuter_rim, R + rc3_rRim, rc3_arOuter_rim_cutout
  );
  #local rc3_RInner_rim_blob = CallwoodIkebanaCSG_Rim_blob_R
  ( rc3_RInner_rim, rc3_arInner_rim,
    rc3_RInner - rc3_rRim, rc3_arInner_rim_cutout
  );

  #local rc3_Rim = union
  {//outer rim:
    CallwoodIkebanaCSG_Torus (rc3_arOuter_rim, rc3_rRim, rc3_ROuter_rim)
   //lavender area (see diagram in documentation):
    difference
    { sphere { 0, rc3_ROuter_rim_blob }
      sphere { 0, R }
      object
      { CallwoodIkebanaCSG_Cone
        ( R, rc3_ROuter_rim_blob, rc3_arOuter_rim_cutout
        )
        inverse
      }
      CallwoodIkebanaCSG_Cone (R, rc3_ROuter_rim_blob + 1, rc3_arOuter_rim)
     //pink area/cutout:
      CallwoodIkebanaCSG_Torus (rc3_arOuter_rim_cutout, rc3_rRim, R + rc3_rRim)
    }
   //inner rim:
    CallwoodIkebanaCSG_Torus (rc3_arOuter_rim, rc3_rRim, rc3_ROuter_rim)
   //lavender area:
    CallwoodIkebanaCSG_Torus (rc3_arInner_rim, rc3_rRim, rc3_RInner_rim)
    difference
    { sphere { 0, rc3_RInner }
      sphere { 0, rc3_RInner_rim_blob }
      object
      { CallwoodIkebanaCSG_Cone
        ( rc3_RInner, rc3_RInner_rim_blob, rc3_arInner_rim_cutout
        )
        inverse
      }
      CallwoodIkebanaCSG_Cone
      ( rc3_RInner + 1, rc3_RInner_rim_blob, rc3_arInner_rim
      )
     //pink area/cutout:
      CallwoodIkebanaCSG_Torus
      ( rc3_arInner_rim_cutout, rc3_rRim, rc3_RInner - rc3_rRim
      )
    }
    rotate <phiHole, 0, 0>
  }

 //--------------- petal assembly -----------------
  #local rc3_Petal = union
  { difference
    { union
      { intersection
        { object { rc3_Full_shell }
          CallwoodIkebanaCSG_Cone (rc3_RInner, R, arPetal - rc3_aMargin)
        }
        intersection
        { object { rc3_Mid_shell }
          CallwoodIkebanaCSG_Cone (rc3_RInner, R, arPetal)
        }
      }
      #local rc3_I = 0;
      #while (rc3_I < 360)
        object { rc3_Hole rotate y * rc3_I }
        #local rc3_I = rc3_I + 120;
      #end
    }
    CallwoodIkebanaCSG_Torus
    ( arPetal - rc3_aMargin, rc3_rOuter_edge, R - rc3_rOuter_edge
    )
    CallwoodIkebanaCSG_Torus
    ( arPetal - rc3_aMargin, rc3_rInner_edge, rc3_RInner + rc3_rInner_edge
    )
    #local rc3_I = 0;
    #while (rc3_I < 360)
      object { rc3_Rim rotate y * rc3_I }
      #local rc3_I = rc3_I + 120;
    #end
   //----- cusp -----
    CallwoodIkebanaCSG_Cusp_layer
    ( rc3_Full_shell, rc3_Cusp_clip,
      object
      { CallwoodIkebanaCSG_Cone (rc3_RInner, R, arCusp + rc3_aMargin)
        transform { rc3_x_Cusp }
      }
    )
    CallwoodIkebanaCSG_Cusp_layer
    ( rc3_Mid_shell, rc3_Cusp_clip,
      object
      { CallwoodIkebanaCSG_Cone (rc3_RInner, R, arCusp)
        transform { rc3_x_Cusp }
      }
    )
    #local rc3_v_Outer = vrotate
    ( (rc3_rOuter_edge - R) * y, rc3_phiCusp * x
    );
    #local rc3_v_Inner = vrotate
    ( -(rc3_RInner + rc3_rInner_edge) * y, rc3_phiCusp * x
    );
    union
    { sphere { rc3_v_Outer, rc3_rOuter_edge }
      sphere { rc3_v_Inner, rc3_rInner_edge }
      #ifdef (CallwoodIkebanaCSG_t_Edge)
        texture { CallwoodIkebanaCSG_t_Edge }
      #end
    }
    #if (!VEq (rc3_v_Outer, rc3_v_Inner))
      Connect_Spheres
      ( rc3_v_Outer, rc3_rOuter_edge, rc3_v_Inner, rc3_rInner_edge
      )
    #end
    object { rc3_Cusp_outer_edge }
    object { rc3_Cusp_outer_edge scale <-1, 1, 1> }
    object { rc3_Cusp_inner_edge }
    object { rc3_Cusp_inner_edge scale <-1, 1, 1> }
    rotate phiPetal * x
  }

 //--------------- neck of petal ------------------
  #local rc3_phiNeck = CallwoodIkebanaCSG_SAdjacent
  ( phiPetal, arPetal + arNeck, rc3_aHalf, (phiPetal < 90? -1: 1),
    "neck of the petals"
  );
  #local rc3_Sanity =
    CallwoodIkebanaCSG_SOpposite (phiPetal, rc3_phiNeck, rc3_aHalf);
  #if (rc3_Sanity != arPetal + arNeck)
    #local rc3_phiNeck2 = 180 - rc3_phiNeck;
    #local rc3_Sanity2 =
      CallwoodIkebanaCSG_SOpposite (phiPetal, rc3_phiNeck2, rc3_aHalf);
    #if (rc3_Sanity2 = arPetal + arNeck)
      #local rc3_phiNeck = rc3_phiNeck2;
    //else both tests failed,
    //  so assume FP error, and stay with smaller (acute) angle.
    #end
  #end
  #local rc3_thetaNeck = CallwoodIkebanaCSG_SAngle
  ( rc3_phiNeck, arNeck + arPetal, phiPetal
  );
  #local rc3_phiBase = CallwoodIkebanaCSG_SOpposite
  ( rc3_phiNeck, arNeck, rc3_thetaNeck
  );
  #local rc3_phiBase_margin = CallwoodIkebanaCSG_SOpposite
  ( rc3_phiNeck, arNeck + rc3_aMargin, rc3_thetaNeck
  );
  #local rc3_Neck_cutout = object
  { CallwoodIkebanaCSG_Cone (rc3_RInner - 1, R + 1, arNeck)
    rotate <rc3_phiNeck, rc3_aHalf, 0>
  }
  #local rc3_Neck_cutout_margin = object
  { CallwoodIkebanaCSG_Cone (rc3_RInner - 1, R + 1, arNeck + rc3_aMargin)
    rotate <rc3_phiNeck, rc3_aHalf, 0>
  }
  #local rc3_Neck_edge = intersection
  { union
    { CallwoodIkebanaCSG_Torus
      ( arNeck + rc3_aMargin, rc3_rInner_edge, rc3_RInner + rc3_rInner_edge
      )
      CallwoodIkebanaCSG_Torus
      ( arNeck + rc3_aMargin, rc3_rOuter_edge, R - rc3_rOuter_edge
      )
    }
    union
    { plane { x, 0 rotate rc3_thetaNeck * y }
      plane { -x, 0 rotate -rc3_thetaNeck * y }
    }
    rotate <rc3_phiNeck, rc3_aHalf, 0>
  }
  #local rc3_Petal_cutout = object
  { CallwoodIkebanaCSG_Cone (rc3_RInner, R, arPetal - rc3_aMargin)
    rotate phiPetal * x
  }

 //--------------- final assembly -----------------
  union
  { difference
    { object { rc3_Mid_shell }
      object { CallwoodIkebanaCSG_Cone (rc3_RInner, R, rc3_phiBase) inverse }
      #local rc3_I = 0;
      #while (rc3_I < nPetals)
        object { rc3_Neck_cutout rotate rc3_I * rc3_aNext * y }
        object { rc3_Petal_cutout rotate rc3_I * rc3_aNext * y }
        #local rc3_I = rc3_I + 1;
      #end
    }
    difference
    { object { rc3_Full_shell }
      object
      { CallwoodIkebanaCSG_Cone (rc3_RInner, R + 1, rc3_phiBase_margin)
        inverse
      }
      #local rc3_I = 0;
      #while (rc3_I < nPetals)
        object { rc3_Neck_cutout_margin rotate rc3_I * rc3_aNext * y }
        object { rc3_Petal_cutout rotate rc3_I * rc3_aNext * y }
        #local rc3_I = rc3_I + 1;
      #end
    }
    #local rc3_I = 0;
    #while (rc3_I < nPetals)
      object { rc3_Neck_edge rotate rc3_I * rc3_aNext * y }
      object { rc3_Petal rotate rc3_I * rc3_aNext * y }
      #local rc3_I = rc3_I + 1;
    #end
  }
#end

//=================================== MATH =====================================

#declare CALLWOODIKEBANACSG_S_ILL_FORMED =
  "The arguments to CallwoodIkebanaCSG() may not describe a well-formed object."

//-------------------- Law of Cosines ----------------------

#macro CallwoodIkebanaCSG_Angle (sideAdj1, sideAdj2, sideOpp)
  acosd
  (   (pow (sideAdj1, 2) + pow (sideAdj2, 2) - pow (sideOpp, 2))
    / (2 * sideAdj1 * sideAdj2)
  )
#end

//--------------- Spherical Law of Cosines -----------------

#macro CallwoodIkebanaCSG_SOpposite (sideAdj1, sideAdj2, Vertex)
  acosd
  (   cosd (sideAdj1) * cosd (sideAdj2)
    + sind (sideAdj1) * sind (sideAdj2) * cosd (Vertex)
  )
#end

// The Sign argument selects between solutions to the quadratic equation.
#macro CallwoodIkebanaCSG_SAdjacent (sideAdj, sideOpp, Vertex, Sign, s_Note)
( #local rc3_Sinb = sind (sideAdj);
  #local rc3_Cosb = cosd (sideAdj);
  #local rc3_Cosc = cosd (sideOpp);
  #local rc3_CosC = cosd (Vertex);
  #local rc3_A = pow (rc3_Sinb * rc3_CosC, 2) + pow (rc3_Cosb, 2);
  #local rc3_B = -2 * rc3_Cosc * rc3_Sinb * rc3_CosC;
  #local rc3_C = pow (rc3_Cosc, 2) - pow (rc3_Cosb, 2);
  #local rc3_Discriminant = pow (rc3_B, 2) - 4 * rc3_A * rc3_C;
  #if (rc3_Discriminant < 0)
    #warning concat
    ( "Negative discriminant ", CallwoodIkebanaCSG_Str (rc3_Discriminant),
      " in the calculation of the ", s_Note, "."
    )
    #if (rc3_Discriminant = 0) //within epsilon, presumably
      #warning "This is likely a floating point rounding error at a right angle, so zero will be substituted."
      #local rc3_Discriminant = 0;
    #else
      #error CALLWOODIKEBANACSG_S_ILL_FORMED
    #end
  #end
  #local rc3_Sine =
    (sqrt (rc3_Discriminant) * (Sign < 0? -1 : 1) - rc3_B) / (2 * rc3_A);
  #if (abs (rc3_Sine) > 1)
    #warning concat
    ( "Inverse sine domain error in the calculation of the ", s_Note, "."
    )
    #error CALLWOODIKEBANACSG_S_ILL_FORMED
  #end
  asind (rc3_Sine)
)
#end

#macro CallwoodIkebanaCSG_SAngle (sideAdj1, sideAdj2, sideOpp)
  acosd
  (   (cosd (sideOpp) - cosd (sideAdj1) * cosd (sideAdj2))
    / (sind (sideAdj1) * sind (sideAdj2))
  )
#end

//================================ AUXILLIARY ==================================

/* A partial cone which, when extended, has its cusp at the origin and pointed
 * upward.
 * Arguments:
 *   rStart, rEnd - the distance from the cusp of the clipping planes; the
 *                  value closer to the origin will be adjusted so that the
 *                  object contains the spherical cap with the given distance
 *                  and angular radius
 *   rAngular     - the angular radius of the cone, in degrees
 */
#macro CallwoodIkebanaCSG_Cone (rStart, rEnd, rAngular)
  #if (rAngular = 90)
    plane { y, 0 }
  #else #if (rAngular > 90)
    object
    { CallwoodIkebanaCSG_Cone (rStart, rEnd, 180 - rAngular)
      scale <1, -1, 1>
      inverse
    }
  #else
    #local rc3_unitR = tand (rAngular);
    #local rc3_rMax = max (rStart, rEnd);
    #local rc3_rMin = min (rStart, rEnd) * cosd (rAngular);
    cone
    { -y * rc3_rMin, rc3_unitR * rc3_rMin, -y * rc3_rMax, rc3_unitR * rc3_rMax
    }
  #end #end
#end

/* A torus for which the central circle lies on a sphere of a given radius.
 * The sphere is centered on the origin, and the torus will be on the bottom.
 * Arguments
 *   aRMajor - the angular major radius, in degrees
 *   rMinor  - the minor radius
 *   RDist   - the radius of the sphere
 */
#macro CallwoodIkebanaCSG_Torus (aRMajor, rMinor, RDist)
  #local rc3_V = CallwoodIkebanaCSG_v_Place_torus (aRMajor, RDist);
  torus
  { rc3_V.x, rMinor translate rc3_V * y
    #ifdef (CallwoodIkebanaCSG_t_Edge)
      texture { CallwoodIkebanaCSG_t_Edge }
    #end
  }
#end

/* Location and major radius of a torus for which the central circle lies on a
 * sphere of a given radius.
 * Arguments
 *   aRMajor - the angular major radius, in degrees
 *   RDist   - the radius of the sphere
 * Returns
 *   A vector with major radius in x and distance below sphere center in y.
 */
#macro CallwoodIkebanaCSG_v_Place_torus (aRMajor, RDist)
  vrotate (-RDist * y, aRMajor * z);
#end

// A cusp of a petal, excluding the rounded edges.
#macro CallwoodIkebanaCSG_Cusp_layer (Shell, Clip, Cutout)
  intersection
  { object { Shell }
    object { Clip }
    object { Cutout inverse }
    object { Cutout inverse scale <-1, 1, 1> }
  }
#end

// Calculates the distance from the origin of the contact between the torus
// that lines the rim of a hole ("rim") and the torus that forms the blob that
// merges it ("cutout") with the surrounding spherical surface.
#macro CallwoodIkebanaCSG_Rim_blob_R (RRim, arRim, RCutout, arCutout)
  vlength
    ((vrotate (RRim * y, arRim * z) + vrotate (RCutout * y, arCutout * z)) / 2)
#end

// A number converted to a string of 6 significant digits, in scientific
// notation if the number is really tiny.
#macro CallwoodIkebanaCSG_Str (Number)
  #if (Number < 0 | Number > 0) //avoid testing against the epsilon
    #local rc3_Log = floor (log (abs(Number)));
  #else
    #local rc3_Log = 0;
  #end
  #if (rc3_Log < -6)
    concat (str (Number * pow (10, -rc3_Log), 1, 5), "E", str (rc3_Log, 1, 0))
  #else
    str (Number, 1, max (5 - rc3_Log, 0))
  #end
#end

//==============================================================================

#version CallwoodIkebanaCSG_Inc_Temp;
#end
